#region LicenseHeader

// Copyright 2012 The Trustees of Indiana University.  All rights reserved.
// 
// The Indiana University licenses this file to you under the Apache License, 
// Version 2.0 (the "License"); you may not use this file except in compliance 
// with the License.  You may obtain a copy of the License at 
// 
// http://www.apache.org/licenses/LICENSE-2.0
// 
// Unless required by applicable law or agreed to in writing, software 
// distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
//  WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  
// See the License for the specific language governing permissions and 
// limitations under the License.
// 
// @author Thilina Gunarathne (tgunarat@indiana.edu)

#endregion

#region

using System;

#endregion

namespace NAligner
{
    /// <summary>
    ///   An implementation of the Smith-Waterman algorithm with Gotoh's improvement for biological local pairwise sequence alignment.
    /// </summary>
    public class SmithWatermanGotoh
    {
        /// <summary>
        ///   Aligns two sequences by Smith-Waterman algorithm
        /// </summary>
        /// <param name="s1"> sequene #1 </param>
        /// <param name="s2"> sequene #2 </param>
        /// <param name="matrix"> scoring matrix </param>
        /// <param name="o"> open gap penalty </param>
        /// <param name="e"> extend gap penalty </param>
        /// <returns> alignment object contains the two aligned sequences, the alignment score and alignment statistics </returns>
        /// <seealso cref="Sequence" />
        /// <seealso cref="Matrix" />
        public PairwiseAlignment AlignDNA(Sequence s1, Sequence s2, float o, float e)
        {
            if (s1 == null)
            {
                Console.WriteLine("Error: S1 is null");

                throw new ArgumentNullException("S1 is null");
            }

            if (s2 == null)
            {
                Console.WriteLine("Error: S2 is null");

                throw new ArgumentNullException("S2 is null");
            }

            int m = s1.Length + 1;
            int n = s2.Length + 1;

            Directions[] pointers = new Directions[m*n];

            // Initializes the boundaries of the traceback matrix to STOP.
            for (int i = 0, k = 0; i < m; i++, k += n)
            {
                pointers[k] = Directions.STOP;
            }

            for (int j = 1; j < n; j++)
            {
                pointers[j] = Directions.STOP;
            }

            short[] sizesOfVerticalGaps = new short[m*n];
            short[] sizesOfHorizontalGaps = new short[m*n];

            for (int i = 0, k = 0; i < m; i++, k += n)
            {
                for (int j = 0; j < n; j++)
                {
                    sizesOfVerticalGaps[k + j] = sizesOfHorizontalGaps[k + j] = 1;
                }
            }

            Cell cell = ConstructDNA(s1, s2, o, e, pointers, sizesOfVerticalGaps, sizesOfHorizontalGaps);

            PairwiseAlignment alignment = TracebackDNA(s1, s2, pointers, cell, sizesOfVerticalGaps,
                                                       sizesOfHorizontalGaps);

            alignment.Name1 = s1.Label;
            alignment.Name2 = s2.Label;
            alignment.Open = o;
            alignment.Extend = e;

            return alignment;
        }

        /// <summary>
        ///   Constructs directions matrix for the traceback
        /// </summary>
        /// <param name="s1"> sequence #1 </param>
        /// <param name="s2"> sequence #2 </param>
        /// <param name="matrix"> scoring matrix </param>
        /// <param name="o"> open gap penalty </param>
        /// <param name="e"> extend gap penalty </param>
        /// <returns> The cell where the traceback starts. </returns>
        private Cell ConstructDNA(Sequence s1, Sequence s2, float o, float e, Directions[] pointers,
                                  short[] sizesOfVerticalGaps, short[] sizesOfHorizontalGaps)
        {
            char[] a1 = s1.Residues.ToCharArray();
            char[] a2 = s2.Residues.ToCharArray();

            int m = s1.Length + 1;
            int n = s2.Length + 1;

            float f; // score of alignment x1...xi to y1...yi if xi aligns to yi
            float[] g = new float[n]; // score if xi aligns to a gap after yi
            float h; // score if yi aligns to a gap after xi
            float[] v = new float[n]; // best score of alignment x1...xi to y1...yi
            float vDiagonal;

            g[0] = float.NegativeInfinity;
            h = float.NegativeInfinity;
            v[0] = 0;

            for (int j = 1; j < n; j++)
            {
                g[j] = float.NegativeInfinity;
                v[j] = 0;
            }

            float similarityScore, g1, g2, h1, h2;

            Cell cell = new Cell();

            for (int i = 1, k = n; i < m; i++, k += n)
            {
                h = float.NegativeInfinity;
                vDiagonal = v[0];

                for (int j = 1, l = k + 1; j < n; j++, l++)
                {
                    // Changed this to avoid the overhead of carring a scoring matrix around each thread or process 
                    // and the additional overhead of making a function call in the trivial case of nucleotide alignments.

                    // begin changes smbeason 6-22-2009 
                    similarityScore = (a1[i - 1] == a2[j - 1]) ? 5 : -4;
                    // end changes smbeason 6-22-2009

                    // Fill the matrices
                    f = vDiagonal + similarityScore;

                    g1 = g[j] - e;
                    g2 = v[j] - o;

                    if (g1 > g2)
                    {
                        // Gap extension penalty
                        g[j] = g1;
                        sizesOfVerticalGaps[l] = (short) (sizesOfVerticalGaps[l - n] + 1);
                    }
                    else
                    {
                        // Gap open penalty
                        g[j] = g2;
                    }

                    h1 = h - e;
                    h2 = v[j - 1] - o;

                    if (h1 > h2)
                    {
                        // Gap extension penalty
                        h = h1;
                        sizesOfHorizontalGaps[l] = (short) (sizesOfHorizontalGaps[l - 1] + 1);
                    }
                    else
                    {
                        // Gap open penalty
                        h = h2;
                    }

                    vDiagonal = v[j];
                    v[j] = Max(f, g[j], h, 0);

                    // Determine the traceback direction
                    if (v[j] == 0)
                    {
                        pointers[l] = Directions.STOP;
                    }
                    else if (v[j] == f)
                    {
                        pointers[l] = Directions.DIAGONAL;
                    }
                    else if (v[j] == g[j])
                    {
                        pointers[l] = Directions.UP;
                    }
                    else
                    {
                        pointers[l] = Directions.LEFT;
                    }

                    // Set the traceback start at the current cell i, j and score
                    if (v[j] > cell.Score)
                    {
                        cell.Set(i, j, v[j]);
                    }
                }
            }

            return cell;
        }

        /// <summary>
        ///   Returns the alignment of two sequences based on the passed array of pointers
        /// </summary>
        /// <param name="s1"> sequence #1 </param>
        /// <param name="s2"> sequence #2 </param>
        /// <param name="m"> scoring matrix </param>
        /// <param name="cell"> The cell where the traceback starts. </param>
        /// <returns> <see cref="Alignment" /> with the two aligned sequences and alignment score. </returns>
        /// <seealso cref="Cell" />
        /// <seealso cref="Alignment" />
        private PairwiseAlignment TracebackDNA(Sequence s1, Sequence s2, Directions[] pointers, Cell cell,
                                               short[] sizesOfVerticalGaps, short[] sizesOfHorizontalGaps)
        {
            char[] a1 = s1.Residues.ToCharArray();
            char[] a2 = s2.Residues.ToCharArray();

            int n = s2.Length + 1;

            PairwiseAlignment alignment = new PairwiseAlignment();
            alignment.Score = cell.Score;

            int maxlen = s1.Length + s2.Length; // maximum length after the aligned sequences

            char[] reversed1 = new char[maxlen]; // reversed sequence #1
            char[] reversed2 = new char[maxlen]; // reversed sequence #2
            char[] reversed3 = new char[maxlen]; // reversed markup

            int len1 = 0; // length of sequence #1 after alignment
            int len2 = 0; // length of sequence #2 after alignment
            int len3 = 0; // length of the markup line

            int identity = 0; // count of identitcal pairs
            int similarity = 0; // count of similar pairs
            int gaps = 0; // count of gaps

            char c1, c2;

            int i = cell.Row; // traceback start row
            int j = cell.Column; // traceback start col
            int k = i*n;

            bool stillGoing = true; // traceback flag: true -> continue & false -> stop

            while (stillGoing)
            {
                switch (pointers[k + j])
                {
                    case Directions.UP:
                        for (int l = 0, len = sizesOfVerticalGaps[k + j]; l < len; l++)
                        {
                            reversed1[len1++] = a1[--i];
                            reversed2[len2++] = PairwiseAlignment.GAP;
                            reversed3[len3++] = Markups.GAP;
                            k -= n;
                            gaps++;
                        }
                        break;

                    case Directions.DIAGONAL:
                        c1 = a1[--i];
                        c2 = a2[--j];
                        k -= n;
                        reversed1[len1++] = c1;
                        reversed2[len2++] = c2;

                        if (c1 == c2)
                        {
                            reversed3[len3++] = Markups.IDENTITY;
                            identity++;
                            similarity++;
                        }
                        else
                        {
                            reversed3[len3++] = Markups.MISMATCH;
                        }
                        break;

                    case Directions.LEFT:
                        for (int l = 0, len = sizesOfHorizontalGaps[k + j]; l < len; l++)
                        {
                            reversed1[len1++] = PairwiseAlignment.GAP;
                            reversed2[len2++] = a2[--j];
                            reversed3[len3++] = Markups.GAP;
                            gaps++;
                        }
                        break;

                    case Directions.STOP:
                        stillGoing = false;
                        break;
                }
            }

            alignment.Sequence1 = Reverse(reversed1, len1);
            alignment.Start1 = i;
            alignment.Sequence2 = Reverse(reversed2, len2);
            alignment.Start2 = j;
            alignment.MarkupLine = Reverse(reversed3, len3);
            alignment.Identity = identity;
            alignment.Gaps = gaps;
            alignment.Similarity = similarity;

            return alignment;
        }

        /// <summary>
        ///   Returns the maximum of 4 float numbers.
        /// </summary>
        private static float Max(float a, float b, float c, float d)
        {
            return Math.Max(Math.Max(a, b), Math.Max(c, d));
        }

        /// <summary>
        ///   Reverses an array of chars
        /// </summary>
        private static char[] Reverse(char[] a, int len)
        {
            // TODO: replace this method by System.Array.Reverse
            char[] b = new char[len];

            for (int i = len - 1, j = 0; i >= 0; i--, j++)
            {
                b[j] = a[i];
            }
            return b;
        }

        #region Old

        /// <summary>
        ///   Aligns two sequences by Smith-Waterman algorithm
        /// </summary>
        /// <param name="s1"> sequene #1 </param>
        /// <param name="s2"> sequene #2 </param>
        /// <param name="matrix"> scoring matrix </param>
        /// <param name="o"> open gap penalty </param>
        /// <param name="e"> extend gap penalty </param>
        /// <returns> alignment object contains the two aligned sequences, the alignment score and alignment statistics </returns>
        /// <seealso cref="Sequence" />
        /// <seealso cref="Matrix" />
        public static PairwiseAlignment Align(Sequence s1, Sequence s2, ScoringMatrix matrix, float o, float e)
        {
            if (matrix == null)
            {
                Console.WriteLine("Error: matrix is null");

                throw new ArgumentNullException("Scoring Matrix is null");
            }

            if (s1 == null)
            {
                Console.WriteLine("Error: S1 is null");

                throw new ArgumentNullException("S1 is null");
            }

            if (s2 == null)
            {
                Console.WriteLine("Error: S2 is null");

                throw new ArgumentNullException("S2 is null");
            }

            float[,] scores = matrix.Scores;

            int m = s1.Length + 1;
            int n = s2.Length + 1;

            Directions[] pointers = new Directions[m*n];

            // Initializes the boundaries of the traceback matrix to STOP.
            for (int i = 0, k = 0; i < m; i++, k += n)
            {
                pointers[k] = Directions.STOP;
            }

            for (int j = 1; j < n; j++)
            {
                pointers[j] = Directions.STOP;
            }

            short[] sizesOfVerticalGaps = new short[m*n];
            short[] sizesOfHorizontalGaps = new short[m*n];

            for (int i = 0, k = 0; i < m; i++, k += n)
            {
                for (int j = 0; j < n; j++)
                {
                    sizesOfVerticalGaps[k + j] = sizesOfHorizontalGaps[k + j] = 1;
                }
            }

            Cell cell = Construct(s1, s2, scores, o, e, pointers, sizesOfVerticalGaps, sizesOfHorizontalGaps);

            PairwiseAlignment alignment = Traceback(s1, s2, matrix, pointers, cell, sizesOfVerticalGaps,
                                                    sizesOfHorizontalGaps);

            alignment.Name1 = s1.Label;
            alignment.Name2 = s2.Label;
            alignment.Matrix = matrix.Name;
            alignment.Open = o;
            alignment.Extend = e;

            return alignment;
        }

        /// <summary>
        ///   Constructs directions matrix for the traceback
        /// </summary>
        /// <param name="s1"> sequence #1 </param>
        /// <param name="s2"> sequence #2 </param>
        /// <param name="matrix"> scoring matrix </param>
        /// <param name="o"> open gap penalty </param>
        /// <param name="e"> extend gap penalty </param>
        /// <returns> The cell where the traceback starts. </returns>
        private static Cell Construct(Sequence s1, Sequence s2, float[,] matrix, float o, float e, Directions[] pointers,
                                      short[] sizesOfVerticalGaps, short[] sizesOfHorizontalGaps)
        {
            char[] a1 = s1.Residues.ToCharArray();
            char[] a2 = s2.Residues.ToCharArray();

            int m = s1.Length + 1;
            int n = s2.Length + 1;

            float f; // score of alignment x1...xi to y1...yi if xi aligns to yi
            float[] g = new float[n]; // score if xi aligns to a gap after yi
            float h; // score if yi aligns to a gap after xi
            float[] v = new float[n]; // best score of alignment x1...xi to y1...yi
            float vDiagonal;

            g[0] = float.NegativeInfinity;
            h = float.NegativeInfinity;
            v[0] = 0;

            for (int j = 1; j < n; j++)
            {
                g[j] = float.NegativeInfinity;
                v[j] = 0;
            }

            float similarityScore, g1, g2, h1, h2;

            Cell cell = new Cell();

            for (int i = 1, k = n; i < m; i++, k += n)
            {
                h = float.NegativeInfinity;
                vDiagonal = v[0];

                for (int j = 1, l = k + 1; j < n; j++, l++)
                {
                    similarityScore = (a1[i - 1] == a2[j - 1]) ? 5 : -4;
                    //similarityScore = matrix[a1[i - 1], a2[j - 1]];

                    // Fill the matrices
                    f = vDiagonal + similarityScore;

                    g1 = g[j] - e;
                    g2 = v[j] - o;

                    if (g1 > g2)
                    {
                        g[j] = g1;
                        sizesOfVerticalGaps[l] = (short) (sizesOfVerticalGaps[l - n] + 1);
                    }
                    else
                    {
                        g[j] = g2;
                    }

                    h1 = h - e;
                    h2 = v[j - 1] - o;

                    if (h1 > h2)
                    {
                        h = h1;
                        sizesOfHorizontalGaps[l] = (short) (sizesOfHorizontalGaps[l - 1] + 1);
                    }
                    else
                    {
                        h = h2;
                    }

                    vDiagonal = v[j];
                    v[j] = Max(f, g[j], h, 0);

                    // Determine the traceback direction
                    if (v[j] == 0)
                    {
                        pointers[l] = Directions.STOP;
                    }
                    else if (v[j] == f)
                    {
                        pointers[l] = Directions.DIAGONAL;
                    }
                    else if (v[j] == g[j])
                    {
                        pointers[l] = Directions.UP;
                    }
                    else
                    {
                        pointers[l] = Directions.LEFT;
                    }

                    // Set the traceback start at the current cell i, j and score
                    if (v[j] > cell.Score)
                    {
                        cell.Set(i, j, v[j]);
                    }
                }
            }

            return cell;
        }

        /// <summary>
        ///   Returns the alignment of two sequences based on the passed array of pointers
        /// </summary>
        /// <param name="s1"> sequence #1 </param>
        /// <param name="s2"> sequence #2 </param>
        /// <param name="m"> scoring matrix </param>
        /// <param name="cell"> The cell where the traceback starts. </param>
        /// <returns> <see cref="Alignment" /> with the two aligned sequences and alignment score. </returns>
        /// <seealso cref="Cell" />
        /// <seealso cref="Alignment" />
        private static PairwiseAlignment Traceback(Sequence s1, Sequence s2, ScoringMatrix m, Directions[] pointers,
                                                   Cell cell, short[] sizesOfVerticalGaps, short[] sizesOfHorizontalGaps)
        {
            char[] a1 = s1.Residues.ToCharArray();
            char[] a2 = s2.Residues.ToCharArray();

            float[,] scores = m.Scores;

            int n = s2.Length + 1;

            PairwiseAlignment alignment = new PairwiseAlignment();
            alignment.Score = cell.Score;

            int maxlen = s1.Length + s2.Length; // maximum length after the aligned sequences

            char[] reversed1 = new char[maxlen]; // reversed sequence #1
            char[] reversed2 = new char[maxlen]; // reversed sequence #2
            char[] reversed3 = new char[maxlen]; // reversed markup

            int len1 = 0; // length of sequence #1 after alignment
            int len2 = 0; // length of sequence #2 after alignment
            int len3 = 0; // length of the markup line

            int identity = 0; // count of identitcal pairs
            int similarity = 0; // count of similar pairs
            int gaps = 0; // count of gaps

            char c1, c2;

            int i = cell.Row; // traceback start row
            int j = cell.Column; // traceback start col
            int k = i*n;

            bool stillGoing = true; // traceback flag: true -> continue & false -> stop

            while (stillGoing)
            {
                switch (pointers[k + j])
                {
                    case Directions.UP:
                        for (int l = 0, len = sizesOfVerticalGaps[k + j]; l < len; l++)
                        {
                            reversed1[len1++] = a1[--i];
                            reversed2[len2++] = PairwiseAlignment.GAP;
                            reversed3[len3++] = Markups.GAP;
                            k -= n;
                            gaps++;
                        }
                        break;

                    case Directions.DIAGONAL:
                        c1 = a1[--i];
                        c2 = a2[--j];
                        k -= n;
                        reversed1[len1++] = c1;
                        reversed2[len2++] = c2;

                        if (c1 == c2)
                        {
                            reversed3[len3++] = Markups.IDENTITY;
                            identity++;
                            similarity++;
                        }
                        else if (scores[c1, c2] > 0)
                        {
                            reversed3[len3++] = Markups.SIMILARITY;
                            similarity++;
                        }
                        else
                        {
                            reversed3[len3++] = Markups.MISMATCH;
                        }
                        break;

                    case Directions.LEFT:
                        for (int l = 0, len = sizesOfHorizontalGaps[k + j]; l < len; l++)
                        {
                            reversed1[len1++] = PairwiseAlignment.GAP;
                            reversed2[len2++] = a2[--j];
                            reversed3[len3++] = Markups.GAP;
                            gaps++;
                        }
                        break;

                    case Directions.STOP:
                        stillGoing = false;
                        break;
                }
            }

            alignment.Sequence1 = Reverse(reversed1, len1);
            alignment.Start1 = i;
            alignment.Sequence2 = Reverse(reversed2, len2);
            alignment.Start2 = j;
            alignment.MarkupLine = Reverse(reversed3, len3);
            alignment.Identity = identity;
            alignment.Gaps = gaps;
            alignment.Similarity = similarity;

            return alignment;
        }

        #endregion
    }
}